"""This subpackage defines the data structures used throughout camfi. Each module makes
heavy use of Pydantic's BaseModel. It is recommended to familiarise yourself with
pydantic before using this library.
"""
